library("data.table")
library("openxlsx")

ot_estimation = function(nominal_catch_data, cpc_data, limit = 5000, reduction_low = .08, reduction_high = .1) {
  ot_catches = nominal_catch_data[GEAR == 'Other']
  ot_catches = merge(ot_catches, cpc_data, by=c("CPC_CODE"))  
  ot_catches[, AVERAGE_CATCH := round(sum(CATCH_T) / 6), by=.(CPC_CODE)] 
  
  CPCs_above_limit = ot_catches[YEAR == 2014 & CATCH_T > limit]
  CPCS_not_subject = ot_catches[AVERAGE_CATCH <= 2000 & !CPC_CODE %in% CPCs_above_limit$CPC_CODE]
  CPCs_below_limit = ot_catches[AVERAGE_CATCH  > 2000 & YEAR == 2014 & !CPC_CODE %in% CPCS_not_subject$CPC_CODE & CATCH_T <= limit]
  
  result = data.table(GEAR=character(), 
                      CPC_CODE=character(), 
                      PARAGRAPH=character(),
                      LIMIT=numeric())
  
  result = rbind(result, ot_estimation_not_subject(ot_catches, CPCS_not_subject))
  result = rbind(result, ot_estimation_above(ot_catches, CPCs_above_limit, reduction_high))
  result = rbind(result, ot_estimation_below(ot_catches, CPCs_below_limit, reduction_low, reduction_high))
  
  result = merge(result, cpc_data, by="CPC_CODE")
  result = merge(result, unique(ot_catches[, .(CPC_CODE, AVERAGE_CATCH)]), by="CPC_CODE")
  result = merge(result, ot_catches[YEAR == 2014, .(CPC_CODE, BASELINE_CATCH = CATCH_T)], all.x = TRUE, by="CPC_CODE")
  result = merge(result, ot_catches[YEAR == 2019, .(CPC_CODE, CATCH_2019 = CATCH_T)], all.x = TRUE, by="CPC_CODE")
  result = result[, .(GEAR, CPC_CODE, CPC, COASTAL, DEV_STATUS, SIDS, BASELINE_YEAR = 2014, BASELINE_CATCH, AVERAGE_CATCH, CATCH_2019, PARAGRAPH, LIMIT)]
  return (result[order(CPC_CODE)])
}

ot_estimation_above = function(nominal_catch_data, cpcs, reduction) {
  result = unique(nominal_catch_data[CPC_CODE %in% cpcs$CPC_CODE, .(GEAR="Other", CPC_CODE, PARAGRAPH="8.a", LIMIT=0.0)])
  
  nominal_catch_data = nominal_catch_data[CPC_CODE %in% cpcs$CPC_CODE & YEAR == 2014]
  
  return(
    result[CPC_CODE %in% nominal_catch_data$CPC_CODE, LIMIT := round((1 - reduction) * nominal_catch_data$CATCH_T)]
  )
}

ot_estimation_below = function(nominal_catch_data, cpcs, reduction_low, reduction_high) {
  result = unique(nominal_catch_data[CPC_CODE %in% cpcs$CPC_CODE, .(GEAR="Other", CPC_CODE, DEV_STATUS, COASTAL, PARAGRAPH="8.b", LIMIT=0.0)])
  
  nominal_catch_data = nominal_catch_data[CPC_CODE %in% cpcs$CPC_CODE & YEAR == 2014]
  
  return(
    result[CPC_CODE %in% nominal_catch_data$CPC_CODE, LIMIT := round(ifelse(DEV_STATUS %in% c("DG", "LD") & COASTAL == TRUE, 1 - reduction_low, 1 - reduction_high) * nominal_catch_data$AVERAGE_CATCH)][, `:=`(DEV_STATUS = NULL, COASTAL = NULL)]
  )
}

ot_estimation_not_subject = function(nominal_catch_data, cpcs) {
  result = unique(nominal_catch_data[CPC_CODE %in% cpcs$CPC_CODE, .(GEAR="Other", CPC_CODE, PARAGRAPH="8.c", LIMIT=0.0)])
  
  nominal_catch_data = nominal_catch_data[CPC_CODE %in% cpcs$CPC_CODE & YEAR == 2019]
  
  result = result[CPC_CODE %in% nominal_catch_data$CPC_CODE, LIMIT := round(nominal_catch_data$CATCH_T)]
  
  return(result)
}

gn_estimation = function(nominal_catch_data, cpc_data, limit = 5000, reduction_low = .12, reduction_high = .15) {
  gn_catches = nominal_catch_data[GEAR == 'Gillnet']
  gn_catches = merge(gn_catches, cpc_data, by=c("CPC_CODE"))  
  gn_catches[, AVERAGE_CATCH := round(sum(CATCH_T) / 6), by=.(CPC_CODE)] 
  
  CPCs_above_limit = gn_catches[YEAR == 2014 & CATCH_T > limit]
  CPCS_not_subject = gn_catches[AVERAGE_CATCH <= 2000 & !CPC_CODE %in% CPCs_above_limit$CPC_CODE]
  CPCs_below_limit = gn_catches[AVERAGE_CATCH  > 2000 & YEAR == 2014 & !CPC_CODE %in% CPCS_not_subject$CPC_CODE & CATCH_T <= limit]
  
  result = data.table(GEAR=character(), 
                      CPC_CODE=character(), 
                      PARAGRAPH=character(),
                      LIMIT=numeric())
  
  result = rbind(result, gn_estimation_not_subject(gn_catches, CPCS_not_subject))
  result = rbind(result, gn_estimation_above(gn_catches, CPCs_above_limit, reduction_high))
  result = rbind(result, gn_estimation_below(gn_catches, CPCs_below_limit, reduction_low, reduction_high))
  
  result = merge(result, cpc_data, by="CPC_CODE")
  result = merge(result, unique(gn_catches[, .(CPC_CODE, AVERAGE_CATCH)]), by="CPC_CODE")
  result = merge(result, gn_catches[YEAR == 2014, .(CPC_CODE, BASELINE_CATCH = CATCH_T)], all.x = TRUE, by="CPC_CODE")
  result = merge(result, gn_catches[YEAR == 2019, .(CPC_CODE, CATCH_2019 = CATCH_T)], all.x = TRUE, by="CPC_CODE")
  result = result[, .(GEAR, CPC_CODE, CPC, COASTAL, DEV_STATUS, SIDS, BASELINE_YEAR = 2014, BASELINE_CATCH, AVERAGE_CATCH, CATCH_2019, PARAGRAPH, LIMIT)]
  return (result[order(CPC_CODE)])
}

gn_estimation_above = function(nominal_catch_data, cpcs, reduction) {
  result = unique(nominal_catch_data[CPC_CODE %in% cpcs$CPC_CODE, .(GEAR="Gillnet", CPC_CODE, PARAGRAPH="6.a", LIMIT=0.0)])
  
  nominal_catch_data = nominal_catch_data[CPC_CODE %in% cpcs$CPC_CODE & YEAR == 2014]
  
  return(
    result[CPC_CODE %in% nominal_catch_data$CPC_CODE, LIMIT := round((1 - reduction) * nominal_catch_data$CATCH_T)]
  )
}

gn_estimation_below = function(nominal_catch_data, cpcs, reduction_low, reduction_high) {
  result = unique(nominal_catch_data[CPC_CODE %in% cpcs$CPC_CODE, .(GEAR="Gillnet", CPC_CODE, DEV_STATUS, COASTAL, PARAGRAPH="6.b", LIMIT=0.0)])
  
  nominal_catch_data = nominal_catch_data[CPC_CODE %in% cpcs$CPC_CODE & YEAR == 2014]
  
  return(
    result[CPC_CODE %in% nominal_catch_data$CPC_CODE, LIMIT := round(ifelse(DEV_STATUS %in% c("DG", "LD") & COASTAL == TRUE, 1 - reduction_low, 1 - reduction_high) * nominal_catch_data$AVERAGE_CATCH)][, `:=`(DEV_STATUS = NULL, COASTAL = NULL)]
  )
}

gn_estimation_not_subject = function(nominal_catch_data, cpcs) {
  result = unique(nominal_catch_data[CPC_CODE %in% cpcs$CPC_CODE, .(GEAR="Gillnet", CPC_CODE, PARAGRAPH="6.c", LIMIT=0.0)])
  
  nominal_catch_data = nominal_catch_data[CPC_CODE %in% cpcs$CPC_CODE & YEAR == 2019]
  
  result = result[CPC_CODE %in% nominal_catch_data$CPC_CODE, LIMIT := round(nominal_catch_data$CATCH_T)]
  
  return(result)
}

ll_estimation = function(nominal_catch_data, cpc_data, limit = 5000, reduction_low = .12, reduction_high = .15) {
  ll_catches = nominal_catch_data[GEAR == 'Longline']
  ll_catches = merge(ll_catches, cpc_data, by=c("CPC_CODE"))  
  ll_catches[, AVERAGE_CATCH := round(sum(CATCH_T) / 6), by=.(CPC_CODE)] 
  
  CPCs_above_limit = ll_catches[YEAR == 2014 & CATCH_T > limit]
  CPCS_not_subject = ll_catches[AVERAGE_CATCH <= 2000 & !CPC_CODE %in% CPCs_above_limit$CPC_CODE]
  CPCs_below_limit = ll_catches[AVERAGE_CATCH  > 2000 & YEAR == 2014 & !CPC_CODE %in% CPCS_not_subject$CPC_CODE & CATCH_T <= limit]
  
  result = data.table(GEAR=character(), 
                      CPC_CODE=character(), 
                      PARAGRAPH=character(),
                      LIMIT=numeric())
  
  result = rbind(result, ll_estimation_not_subject(ll_catches, CPCS_not_subject))
  result = rbind(result, ll_estimation_above(ll_catches, CPCs_above_limit, reduction_high))
  result = rbind(result, ll_estimation_below(ll_catches, CPCs_below_limit, reduction_low, reduction_high))
  
  result = merge(result, cpc_data, by="CPC_CODE")
  result = merge(result, unique(ll_catches[, .(CPC_CODE, AVERAGE_CATCH)]), by="CPC_CODE")
  result = merge(result, ll_catches[YEAR == 2014, .(CPC_CODE, BASELINE_CATCH = CATCH_T)], all.x = TRUE, by="CPC_CODE")
  result = merge(result, ll_catches[YEAR == 2019, .(CPC_CODE, CATCH_2019 = CATCH_T)], all.x = TRUE, by="CPC_CODE")
  result = result[, .(GEAR, CPC_CODE, CPC, COASTAL, DEV_STATUS, SIDS, BASELINE_YEAR = 2014, BASELINE_CATCH, AVERAGE_CATCH, CATCH_2019, PARAGRAPH, LIMIT)]
  return (result[order(CPC_CODE)])
}

ll_estimation_above = function(nominal_catch_data, cpcs, reduction) {
  result = unique(nominal_catch_data[CPC_CODE %in% cpcs$CPC_CODE, .(GEAR="Longline", CPC_CODE, PARAGRAPH="7.a", LIMIT=0.0)])
  
  nominal_catch_data = nominal_catch_data[CPC_CODE %in% cpcs$CPC_CODE & YEAR == 2014]
  
  return(
    result[CPC_CODE %in% nominal_catch_data$CPC_CODE, LIMIT := round((1 - reduction) * nominal_catch_data$CATCH_T)]
  )
}

ll_estimation_below = function(nominal_catch_data, cpcs, reduction_low, reduction_high) {
  result = unique(nominal_catch_data[CPC_CODE %in% cpcs$CPC_CODE, .(GEAR="Longline", CPC_CODE, DEV_STATUS, COASTAL, PARAGRAPH="7.b", LIMIT=0.0)])
  
  nominal_catch_data = nominal_catch_data[CPC_CODE %in% cpcs$CPC_CODE & YEAR == 2014]
  
  return(
    result[CPC_CODE %in% nominal_catch_data$CPC_CODE, LIMIT := round(ifelse(DEV_STATUS %in% c("DG", "LD") & COASTAL == TRUE, 1 - reduction_low, 1 - reduction_high) * nominal_catch_data$AVERAGE_CATCH)][, `:=`(DEV_STATUS = NULL, COASTAL = NULL)]
  )
}

ll_estimation_not_subject = function(nominal_catch_data, cpcs) {
  result = unique(nominal_catch_data[CPC_CODE %in% cpcs$CPC_CODE, .(GEAR="Longline", CPC_CODE, PARAGRAPH="7.c", LIMIT=0.0)])
  
  nominal_catch_data = nominal_catch_data[CPC_CODE %in% cpcs$CPC_CODE & YEAR == 2019]
  
  result = result[CPC_CODE %in% nominal_catch_data$CPC_CODE, LIMIT := round(nominal_catch_data$CATCH_T)]
  
  return(result)
}

ps_estimation = function(nominal_catch_data, cpc_data, limit = 5000) {
  ps_catches = nominal_catch_data[GEAR == 'Purse seine']
  ps_catches = merge(ps_catches, cpc_data, by=c("CPC_CODE"))  
  ps_catches[, AVERAGE_CATCH := round(sum(CATCH_T) / 6), by=.(CPC_CODE)] 

  CPCs_above_limit = ps_catches[YEAR == PS_BASELINE_YEAR & CATCH_T > limit]
  CPCS_not_subject = ps_catches[AVERAGE_CATCH <= 2000 & !CPC_CODE %in% CPCs_above_limit$CPC_CODE]
  CPCs_below_limit = ps_catches[AVERAGE_CATCH  > 2000 & YEAR == PS_BASELINE_YEAR & !CPC_CODE %in% CPCS_not_subject$CPC_CODE & CATCH_T <= limit]

  result = data.table(GEAR=character(), 
                      CPC_CODE=character(), 
                      PARAGRAPH=character(),
                      LIMIT=numeric())
  
  result = rbind(result, ps_estimation_not_subject(ps_catches, CPCS_not_subject))
  result = rbind(result, ps_estimation_above(ps_catches, CPCs_above_limit))
  result = rbind(result, ps_estimation_below(ps_catches, CPCs_below_limit))
  
  result = merge(result, cpc_data, by="CPC_CODE")
  result = merge(result, unique(ps_catches[, .(CPC_CODE, AVERAGE_CATCH)]), by="CPC_CODE")
  result = merge(result, ps_catches[YEAR == PS_BASELINE_YEAR, .(CPC_CODE, BASELINE_CATCH = CATCH_T)], all.x = TRUE, by="CPC_CODE")
  result = merge(result, ps_catches[YEAR == 2019, .(CPC_CODE, CATCH_2019 = CATCH_T)], all.x = TRUE, by="CPC_CODE")
  result = result[, .(GEAR, CPC_CODE, CPC, COASTAL, DEV_STATUS, SIDS, BASELINE_YEAR = PS_BASELINE_YEAR, BASELINE_CATCH, AVERAGE_CATCH, CATCH_2019, PARAGRAPH, LIMIT)]
  return (result[order(CPC_CODE)])
}

ps_estimation_above = function(nominal_catch_data, cpcs) {
  result = unique(nominal_catch_data[CPC_CODE %in% cpcs$CPC_CODE, .(GEAR="Purse seine", CPC_CODE, SIDS, PARAGRAPH="5.a", LIMIT=0.0)])
  
  nominal_catch_data = nominal_catch_data[CPC_CODE %in% cpcs$CPC_CODE & YEAR == PS_BASELINE_YEAR]

  result = result[CPC_CODE %in% nominal_catch_data$CPC_CODE, LIMIT := round(ifelse(SIDS == TRUE, 1 - 0.16, 1 - 0.2) * nominal_catch_data$CATCH_T)]
  result[SIDS == TRUE, PARAGRAPH := "5.d"]
  
  return(result[, SIDS := NULL])
}

ps_estimation_below = function(nominal_catch_data, cpcs) {
  result = unique(nominal_catch_data[CPC_CODE %in% cpcs$CPC_CODE, .(GEAR="Purse seine", CPC_CODE, DEV_STATUS, COASTAL, PARAGRAPH="5.b", LIMIT=0.0)])
  
  nominal_catch_data = nominal_catch_data[CPC_CODE %in% cpcs$CPC_CODE & YEAR == PS_BASELINE_YEAR]
  
  return(
    result[CPC_CODE %in% nominal_catch_data$CPC_CODE, LIMIT := round(ifelse(DEV_STATUS %in% c("DG", "LD") & COASTAL == TRUE, 1 - 0.17, 1 - 0.2) * nominal_catch_data$AVERAGE_CATCH)][, `:=`(DEV_STATUS = NULL, COASTAL = NULL)]
  )
}

ps_estimation_not_subject = function(nominal_catch_data, cpcs) {
  result = unique(nominal_catch_data[CPC_CODE %in% cpcs$CPC_CODE, .(GEAR="Purse seine", CPC_CODE, PARAGRAPH="5.c", LIMIT=0.0)])
  
  nominal_catch_data = nominal_catch_data[CPC_CODE %in% cpcs$CPC_CODE & YEAR == 2019]
    
  result = result[CPC_CODE %in% nominal_catch_data$CPC_CODE, LIMIT := round(nominal_catch_data$CATCH_T)]
  
  return(result)
}

produce_estimates = function(input_scenario = "CIRCULAR_2020_55", mnemonic = "CIR") {
  DATA_FOLDER = "../data/"
  INPUT_FOLDER = paste0(DATA_FOLDER, "input/scenario/")
  OUTPUT_FOLDER = paste0(DATA_FOLDER, "output/")

  CPCs = data.table(read.xlsx(paste0(DATA_FOLDER, "reference/CPC_list.xlsx")))
  
  NCRAW = data.table(read.xlsx(paste0(INPUT_FOLDER, input_scenario, "/", mnemonic, "_YFT_2014_2019.xlsx"), sheet = 'DATA'))
  NCRAW[, CPC := NULL]
  NCRAW[CPC_CODE == 'TWN', `:=` (CPC_CODE='CHN')]
  NCRAW[GEAR %in% c('Baitboat','Line'), GEAR := 'Other']
  
  NC = NCRAW[, .(CATCH_T=sum(CATCH_T)),keyby=.(YEAR,GEAR,CPC_CODE)]
  
  PS = ps_estimation(NC, CPCs)
  
  LL_MDV = ll_estimation(NC, CPCs, 5000, 0.12, 0.15)
  LL_EU  = ll_estimation(NC, CPCs, 5000, 0.17, 0.20)
  
  GN_MDV = gn_estimation(NC, CPCs, 5000, 0.12, 0.15)
  GN_EU  = gn_estimation(NC, CPCs, 2000, 0.17, 0.20)
  
  OT = ot_estimation(NC, CPCs, 5000, 0.08, 0.10)
  
  ALL_MDV = rbindlist(list(PS, LL_MDV, GN_MDV, OT))
  ALL_EU  = rbindlist(list(PS, LL_EU , GN_EU , OT))
  
  CPC_LIMIT_MDV = ALL_MDV[, .(TOTAL_LIMIT=sum(LIMIT)), keyby = .(CPC_CODE, CPC, COASTAL, DEV_STATUS, SIDS)]
  CPC_LIMIT_MDV[, MAX_LIMIT := ifelse(COASTAL == TRUE & TOTAL_LIMIT < 2000, 2000, TOTAL_LIMIT)]
  
  CPC_LIMIT_EU  = ALL_EU[, .(TOTAL_LIMIT=sum(LIMIT)), keyby = .(CPC_CODE, CPC, COASTAL, DEV_STATUS, SIDS)]
  CPC_LIMIT_EU [, MAX_LIMIT := ifelse(COASTAL == TRUE & TOTAL_LIMIT < 2000, 2000, TOTAL_LIMIT)]
  
  write.xlsx(PS, paste0(OUTPUT_FOLDER, input_scenario, "/EU/", mnemonic, "_PS_EU.xlsx"))
  write.xlsx(PS, paste0(OUTPUT_FOLDER, input_scenario, "/MDV/", mnemonic, "_PS_MDV.xlsx"))
 
  write.xlsx(LL_EU, paste0(OUTPUT_FOLDER, input_scenario, "/EU/", mnemonic, "_LL_EU.xlsx"))
  write.xlsx(LL_MDV, paste0(OUTPUT_FOLDER, input_scenario, "/MDV/", mnemonic, "_LL_MDV.xlsx"))
  
  write.xlsx(GN_EU, paste0(OUTPUT_FOLDER, input_scenario, "/EU/", mnemonic, "_GN_EU.xlsx"))
  write.xlsx(GN_MDV, paste0(OUTPUT_FOLDER, input_scenario, "/MDV/", mnemonic, "_GN_MDV.xlsx"))
  
  write.xlsx(OT, paste0(OUTPUT_FOLDER, input_scenario, "/EU/", mnemonic, "_OT_EU.xlsx"))
  write.xlsx(OT, paste0(OUTPUT_FOLDER, input_scenario, "/MDV/", mnemonic, "_OT_MDV.xlsx"))
  
  write.xlsx(ALL_EU, paste0(OUTPUT_FOLDER, input_scenario, "/EU/", mnemonic, "_ALL_EU.xlsx"))
  write.xlsx(ALL_MDV, paste0(OUTPUT_FOLDER, input_scenario, "/MDV/", mnemonic, "_ALL_MDV.xlsx"))
  
  ALL_EU$MEMBER = "PROP B (EU)"
  ALL_MDV$MEMBER = "PROP C (MDV)"
  
  ALL = rbind(ALL_EU, ALL_MDV)
  
  write.xlsx(ALL, paste0(OUTPUT_FOLDER, input_scenario, "/", mnemonic, "_ALL.xlsx"))
  
  write.xlsx(CPC_LIMIT_EU, paste0(OUTPUT_FOLDER, input_scenario, "/EU/", mnemonic, "_TOTAL_LIMITS_EU.xlsx"))
  write.xlsx(CPC_LIMIT_MDV, paste0(OUTPUT_FOLDER, input_scenario, "/MDV/", mnemonic, "_TOTAL_LIMITS_MDV.xlsx"))

  CPC_LIMIT_EU$MEMBER = "PROP B (EU)"
  CPC_LIMIT_MDV$MEMBER = "PROP C (MDV)"
  
  CPC_LIMIT = rbind(CPC_LIMIT_EU, CPC_LIMIT_MDV)
  
  write.xlsx(CPC_LIMIT, paste0(OUTPUT_FOLDER, input_scenario, "/", mnemonic, "_TOTAL_LIMITS_ALL.xlsx"))
}

produce_estimates("BEST_SCIENTIFIC_ESTIMATES", "SCI")